/**
 *  The Main Library of Betterville
 *  @author Drakkos
 *  @started 26th of March, 2009
 *
 */
 
#include "path.h"
// We need the library.h include for access to the quest library.
#include <library.h>
// We need tasks.h for access to the taskmaster.
#include <tasks.h>

// Define these so that we don't need to worry about spelling when 
// working with the library.
#define QUEST_NAME "betterville library"
#define HOLE_QUEST "betterville hole"

// Make use of the inerit that we ourselves wrote.
inherit INHERITS + "inside_room";

// We create our own custom data type here - the 'book'.  Each of 
// the variables within the coding block here refers to a 'compartment' 
// in this variable.
class book {
  string title;
  string author;
  string blurb;
  string *categories;
}
// We set up an array of our books - rather than just saying 
// 'book *allBooks' as we would with 'int *allBooks' or such,
// we need to explicitly note that it is a class to which we
// are referring.
class book *allBooks = ({ });

// This mapping contains the relationship of book to sorted 
// category.  It's the state of the library as sorted by the 
// player, rather than the state of the library as it should 
// be. 
mapping sorting;

// The usual function prototypes.  Remember, we need to prototype 
// any function that we use before it gets declared in the code.
void reset_sorting();
void setup_quest();
string books_to_sort();
string books_on_shelves();
string book_categories();
int find_book (string book);
string *query_categories();
string query_book_blurb (string);
int assign_book_to_shelf (string, string);
int check_complete ();
string* check_hole_location (object);

/**
 *  
 *  This method takes an array of strings and returns it in a pretty 
 *  format suitable for presentation.  Each element of the array is 
 *  capitalised, and surrounded in quotation marks.
 *
 *  @param    arr       The array to make pretty
 *  @returns            A pretty string for output
 *
 */
 
string pretty_array_output (string *arr) {
  for (int i = 0; i < sizeof (arr); i++) {
    arr[i] = "\"" + cap_words (arr[i]) + "\"";
  }
  // query_multiple_short is an sfun, and it takes an array of 
  // objects or strings and returns them in the form 'A, B, and C' 
  return query_multiple_short (arr);
}

/**
 *
 *  This method generates a random code for the player. 
 *
 *  @returns            A random code to be used in a later quest
 *
 */
 
string generate_random_code() {
  int rand = random (9000);
  rand += 1000;
 
  // We add the random number to an empty string - this forces a type 
  // conversion from the number 1234 to the string "1234" 
  return "" + rand;
}

void setup() {
  set_short ("main library");
  add_property ("determinate", "the ");
  set_long ("This is the main part of the library.  It's a mess, with "
    "discarded books thrown everywhere!\n");
  set_light (100);
  
  // These add_items each point to a function - the function returns the text 
  // to be displayed when the user looks at the item.
  add_item ("book", (: books_to_sort() :));
  add_item (({"shelf", "bookshelf"}), (: books_on_shelves() :));
  add_item ("category", (: book_categories() :));
  
  add_exit ("west", ROOMS + "betterville_09", "door");

  // We call the methods to set the initial starting state of our quest.
  setup_quest();
  // We reset the sorting mapping so that the library is completely unsorted.
  reset_sorting();
}

void init() {
  ::init();
  // Holy smokes, four commands!  We need these to manage the user interaction 
  // with our quest.  The first part of add_command is the command we provide to 
  // the player, and the second part is the pattern with which we match player 
  // input.  When a valid combination of command and pattern is provided by 
  // the player, the method do_<name_of_command> is called.

  // Calls the method do_sort
  add_command ("sort", "<string'book'> into [category] <string'category'>");
  // Calls the method do_read
  add_command ("read", "blurb on <string'book'>");
  // Calls the method to_break
  add_command ("break", "roof above <string'category'> bookshelf");
  // Calls the method do_climb
  add_command ("climb", "through roof");
}

int do_climb (object *indirect_obs, string dir_match, string indir_match, 
  mixed *args, string pattern) {

  // Check to see if they have completed the pre-requisite quest.  It would 
  // be tedious for people to have to do the same task every time they wanted 
  // to get to the second floor of the library, so we just base it on their 
  // having previously completed the neccessary steps.
  if (!LIBRARY->query_quest_done (this_player()->query_name(), HOLE_QUEST)) {
    add_failed_mess ("How are you going to manage that, you crazy diamond?\n",
      ({ }));
    return 0;
  }

  // Put this through the taskmaster, just for the hell of it.  Really, we 
  // don't have to, but why not?
  switch(TASKER->perform_task(this_player(), "adventuring.movement.climbing.rope", 
    100, TM_FREE ) ) {
      case AWARD :
        tell_object(this_player(), "%^YELLOW%^%^BOLD%^You feel a little more "
          "nimble.%^RESET%^\n" );
      case SUCCEED :
        this_player()->move_with_look (ROOMS + "library_second_level",
          "$N arrive$s through a hole in the roof.", "$N climb$s up through "
          "a hole in the roof.");
        tell_object (this_player(), "You climb up through the hole in the "
          "roof you created.\n");
        return 1;
        break;
      default :
        add_succeeded_mess ("$N attempt$s to climb through the roof, but "
          "fail$s horribly.\n", ({ }));
        return 1;
  }
  
}
int do_break (object *indirect_obs, string dir_match, string indir_match, 
  mixed *args, string pattern) {

  // The args array holds each of the 'fill in the blanks' values that 
  // the player gave us.  We know the first element in the array then 
  // is going to be the category bookshelf over which the weak part in 
  // the roof can be seen.
  string category = args[0];
  string *details;
  string location, tool;
  object *valid_tools;

  // First, we want to see if they've done the quest before.  If they have, 
  // let them know this isn't what they have to do each time.  
  if (LIBRARY->query_quest_done (this_player()->query_name(), HOLE_QUEST)) {
    add_failed_mess ("You have already broken through the roof, no need to "
      "do it again.\n", ({ }));
    return 0;
  }
  
  // Check hole location gives us back an array of strings.  The first of these 
  // is the location of the hole, the second is the tool we want to use.  We could 
  // do this is a number of different ways, such as returning a class or a mapping, 
  // but this works and is simple enough.
  details = check_hole_location (this_player());
  
  // If details holds 0, then they have discovered no hole to break.
  if (!details) {
    add_failed_mess ("Break what?  Why?  You are a bad person for being such "
      "a vandal.\n", ({ }));
    return 0;
  }
  
  // Get the location where the hole is.
  location = details[0];
  // Get the tool they need to break through
  tool = details[1];
  
  // Check to see if they're breaking through at the part where they saw the 
  // weak roof.
  if (category != location) {
    add_failed_mess ("The roof there is too solid to break.\n", ({ }));
    return 0;
  }
  
  // Okay, now we look to see if they have the right kind of tool in their 
  // inventory.  We could do all sorts of things here like checking to see 
  // if they are actually holding it or such, but I find that just tedious 
  // busywork when trying to accomplish a task in the game.  Let's assume 
  // they are bright enough to be able to use a tool in their inventory
  // rather than have to spell it out.
  valid_tools = match_objects_for_existence (tool, this_player(), 
    this_player());
     
  // If they have no valid tools, they fail.
  if (!sizeof (valid_tools)) {
    add_failed_mess ("You have no tool suitable for breaking through the "
      "roof.\n", ({ }));
    return 0;
  }
  
  // Huzzah, they broke through the roof at the right place and they 
  // got the quest!
  LIBRARY->set_quest (this_player()->query_name(), HOLE_QUEST);
  
  // add_succeeded_mess gives us a way of giving the proper output to 
  // the room.  $N gets parsed into the player name, break$s is a pluralisation 
  // helper (it will display as 'break' to the player and 'breaks' to everyone 
  // else in the room).  $I represents an indirect object, and we provide the 
  // list of these as the array parameter.  They get parsed with query_multiple_short.
  add_succeeded_mess ("$N break$s through the roof with $I.\n", 
    ({ valid_tools[0] }));
  return 1;
}

int do_read (object *indirect_obs, string dir_match, string indir_match, 
  mixed *args, string pattern) {
  // Get the book that they entered into the command.
  string book = lower_case (args[0]);
  string blurb;
  
  // Get the blurb for the book they indicated.
  blurb = query_book_blurb (book);
  
  // If there's no blurb, then there's no book.
  if (!blurb) {
    add_failed_mess ("That book does not seem to exist.\n");
    return 0;
  }
  
  // We tell them what the blurb is.  To ensure that it gets mangled 
  // properly through known languages and skills, we use the $L$ 
  // coding system.  
  tell_object (this_player(), "The blurb on the book reads: " + 
    "$L$[read:morporkian]" + blurb + "$L$\n");
 
 // Display the appropriate message to the room and the player.
  add_succeeded_mess ("$N pick$s up a book and read$s the back.\n",
    ({ }));
  return 1;
}

int do_sort (object *indirect_obs, string dir_match, string indir_match, 
  mixed *args, string pattern) {

  string code;
  // Get the book they entered.
  string book = lower_case (args[0]);
  // Get the category they entered.
  string category = lower_case (args[1]);

  // If they have already done the quest, stop them trying to do it again.
  if (LIBRARY->query_quest_done (this_player()->query_name(), QUEST_NAME)) {
    add_failed_mess ("You have already done this quest.  Give someone else a "
      "chance.\n", ({ }));
    return 0;
  }

  // Check to see if the library is already organised.  No need to mess it 
  // up if it is.
  if (check_complete()) {
    add_failed_mess ("The library is immaculately organised.  Don't go "
      "spoiling it now.\n", ({ }));
    return 0;
  }
  
  // Check to see if the book they entered actually exists.
  if (find_book (book) == -1) {
    add_failed_mess ("That book does not seem to exist.\n", ({ }));
    return 0;
  }
  
  // Check to see if the category they entered actually exists.
  if (member_array (category, query_categories()) == -1) {
    add_failed_mess ("That category does not seem to exist.\n", ({ }));
    return 0;
  }
  
  // Both the book and the category exist, so make their actions known.
  assign_book_to_shelf (book, category);
  
  // Now that we've changed the state of the library, let's see if it's 
  // now in a sorted state.
  if (check_complete()) {
    // It is, so we set the player as having done the quest.
    LIBRARY->set_quest (this_player()->query_name(), QUEST_NAME);
    // We also generate a random code for them as set their quest info 
    // for this quest appropriately.
    LIBRARY->set_player_quest_info (this_player()->query_name(), QUEST_NAME, 
      generate_random_code());    
  
    // We could do this a little more concisely, but this is proof of concept.  
    // To show that we actually stored the information we want, we pull it 
    // back out of the quest handler.
    code = LIBRARY->query_player_quest_info (this_player()->query_name(), 
      QUEST_NAME);
    
    // We show the player the random code that is associated with their 
    // quest.
    tell_object (this_player(), "As you put away the last book, a small slip "
      "of aged paper flutters to the ground.  You can see it has the numbers \"" + 
      code + "\" on it before it crumbles into dust.\n");
  }
  
  // Show the appropriate message to the room.
  add_succeeded_mess ("$N sort$s the book \"" + cap_words (book) + "\" into category \""  + 
    cap_words (category) + "\"\n", ({ }));
  return 1;
}

void reset() {
  sorting = ([ ]);
}

/**
 *
 *  Add a book to the allBooks array.  Doing this will generate a random category for 
 *  the book, as well as a category specific blurb for it.
 *
 *  @param    book_title      The name of the book to add
 *  @param    book_author     The author of the book to add
 *
 */
 
void add_book(string book_title, string book_author) {
  // This is the container for the new book we are going to create and 
  // then add to the allBooks array.
  class book newBook;
  string my_blurb;
  // Setup what categories we have as valid.
  string *cats = ({"romance", "action", "childrens"});
  string my_cat;
  string *valid_blurbs;  
  // Setup the valid blurbs.
  mapping blurbs = ([
    "romance" : 
    ({
        "A romantic tale of two estranged lovers",
        "A heartbreaking tale of forbidden love!",
        "A story of two monkeys, with only each other to rely on!",
    }),
    "action" : 
    ({
        "A thrilling adventure complete with a thousand elephants!",
        "An exciting story full of carriages-chases and brutal sword-fights!",
        "A story of bravery and heroism in the trenches of Koom Valley!",
    }),
    "childrens":
    ({
        "A heart-warming tale of a fuzzy family cat!",
        "A story about a lost cow, for children!",
        "A educational story about Whiskerton Meowington and his last trip to the vet!",
    })
  ]);
  
  // Make the book title lower_case for consistency purposes.
  book_title = lower_case (book_title);
  
  // Pick a random category
  my_cat = element_of (cats);
  // Get the blurbs that go with that category
  valid_blurbs = blurbs[my_cat];
  // Pick a random blurb from the valid blurbs
  my_blurb = element_of (valid_blurbs);
  
  // Create the instance of the book object from the information we have.
  newBook = new (class book, 
    title: book_title,
    author: book_author,
    blurb: my_blurb,
    categories: ({my_cat}));

  // Add that  new book object to the allBooks array.
  allBooks += ({ newBook });
}

/**
 *
 *  This method gets a list of all the categories of books in the 
 *  library.
 *
 *  @returns        The list of categories
 *
 */
 
string *query_categories() {
  string *cats = ({ });
  
  for (int i = 0 ; i < sizeof (allBooks); i++) {
    foreach (string c in allBooks[i]->categories) {
      if (member_array (c, cats) == -1) {
        cats += ({ c });
      }
    }
  }
  
  return cats;
}


/**
 *
 *  This method sets up the starting state of the quest.
 *
 */
 
void setup_quest() {
  add_book ("Some stuff about you", "you");
}

/**
 *  This method finds a book from the allBooks array.  It returns -1 if the 
 *  book does not exist, and the index of the book in the array if it does.
 *
 *  @param      title       The name of the book to find
 *  @returns                The index of the book in the array, or -1 if it
 *                          doesn't exist.
 *
 */
 
int find_book (string title) {
  for (int i = 0; i < sizeof (allBooks); i++) {
    if (allBooks[i]->title == title) {
      return i;
    }
  }
  return -1;
}

/**
 *
 *  Take a book and a category, and mark it as being sorted.
 *
 *  @param      book        The book to sort
 *  @param      category    The category into which the book should be sorted
 *  @returns                -1 if no valid book, -2 if no valid category, 1 
 *                          if successful
 *
 */
 
int assign_book_to_shelf (string book, string category) {
  int i = find_book (book);
  string *cats;
  
  if (i == -1) {
    return -1;
  }
  
  cats = query_categories();
  
  if (member_array (category, cats) == -1) {
    return -2;
  }
  
  sorting[book] = category;
  return 1;
}

/**
 *  
 *  Query the blurb associated with a particular book.
 *
 *  @param      title     The name of the book to get the blurb for
 *  @returns              The blurb of the book, or 0 if th ebook doesn't exist
 *
 */
 
string query_book_blurb (string title) {
  int i = find_book (title);
  
  if (i == -1) {
    return 0;
  }
  
  return allBooks[i]->blurb;
}

/**
 *
 *  Get the list of all books that have not yet been sorted into categories.
 *
 *  @returns              The list of unsorted books
 *
 */
 
string *query_unsorted_books() {
  string *unsorted = ({ });
  
  for (int i = 0; i < sizeof (allBooks); i++) {
    if (!sorting[allBooks[i]->title]) {
      unsorted += ({ allBooks[i]->title });
    }
  }
  
  return unsorted;
}

/**
 *
 *  Get the bookshelf to which a particular book has been assigned.
 *
 *  @param      title     The book to get the shelf of
 *  @returns              The shelf the book has been sorted to, or 0 if it 
 *                        does not exist or the book has not been sorted.
 *
 */
  
string query_bookshelf_of_book (string title) {
  return sorting[title];
}

/**
 *  
 *  Reset the state of the sorting mapping
 *
 */
 
void reset_sorting() {
  sorting = ([ ]);
}

/**
 *
 *  This method gets the text associated with the books add_item
 *
 *  @returns          the add_item text
 *
 */
 
string books_to_sort() {
  // First, we get a list of unsorted books.
  string *unsorted = query_unsorted_books();
  
  // if there are no unsorted books, then the library has been sorted.  Perhaps not
  // correctly, but they've all been shelved.
  if (!sizeof (unsorted)) {
    return "All the books have been neatly sorted away.";
  }
  else {
    // Give a list of the books remaining to be sorted, in a user friendly way.
    return "The following books are strewn around the library: " + 
      pretty_array_output (unsorted) + ".  They are just begging to "
      "be given a good sorting.  Each has a blurb on the back that "
      "may help in discovering to what category they belong.";
  }  
}

/**
 *
 *  This method gets the text associated with the shelves add_item
 *
 *  @returns          the add_item text
 *
 */

string books_on_shelves() {
  // This is the first part of our add_item text.   We'll add more specialised 
  // details onto this as we go through the function.
  string ret = "The bookshelves are arranged according to categories.  ";
  string *details;
  
  // Get the details of the hole location.  We're going to bind noticing 
  // that into this add_item.
  details = check_hole_location (this_player());
  
  // If no books have been sorted, 
  if (!sorting || !sizeof (sorting)) {
    // Add this to our first part of the text.
    ret += "No books have been sorted onto the shelves.";
  }
  else {
    // Go over each of the sorted books.
    foreach (string book, string category in sorting) {
      // Add the specific details of this book to the add_item
      ret += "The book \"" + cap_words (book) + "\" has been put on the shelf "
        "for the category \"" + cap_words (category) + "\"\n";
    }
  }
  
  // If there are some details returned from check_hole_location, reveal them as part of 
  // the add_item.
  if (details) {
    ret += "  You can see a crack in the roof above the \"" + details[0] 
      + "\" bookshelf.  If you had a " + details[1] + ", you could probably "
      "break through the roof.";
  }
  
  return ret;
}

/**
 *
 *  This method gets the text associated with the category add_item
 *
 *  @returns          the add_item text
 *
 */

string book_categories() {
  return "The following categories are labelled on the bookshelves: " + 
    pretty_array_output (query_categories());
}

/**
 *
 *  This method checks to see if the current state of the library matches the 
 *  desired state of the library.  The desired state is when each of the books
 *  is sorted into a category marked as valid.
 *
 *  @returns            1 if the library is sorted, 0 if it is not.
 *
 */
 
int check_complete() {
  int i;
 
  // If there are books remaining to be sorted, then the sorting is not 
  // complete. 
  if (sizeof (query_unsorted_books())) {
    return 0;
  }
  
  // Step over each of the books that have been sorted.
  foreach (string book, string category in sorting) {
    // Find that book in the array.
    i = find_book (book);
    
    // This shouldn't happen!  But let's make sure if it does, our code still 
    // soldiers on.
    if (i == -1) {
      return 0;
    }
    
    // Okay, now check to see whether or not the category the book was sorted 
    // into matches one of the categories belonging to that book.
    if (member_array (category, allBooks[i]->categories) == -1) {
      // If the answer for any book is 'no', then the library is not sorted
      // properly.
      return 0;
    }
  }
  
  // No problems with the books, so the library is sorted.
  return 1;
}

/**
 *
 *  This is the method that handles whether or not a player has noticed a 
 *  hole in the roof.  It handles the taskmaster check and storing the quest
 *  data. 
 *
 *  @param        player        The player in whom we are interested
 *  @returns                    0 if the player has noticed nothing, and an array
 *                              containing the location and the tool needed if 
 *                              they have.
 *
 */
   
string* check_hole_location (object player) {
  // Okay, let's pull their details for this quest out of the library.
  string *details = LIBRARY->query_player_quest_info (player->query_name(), 
    HOLE_QUEST);
  // Setup the list of categories where a hole may be noticed.
  string *cat = ({"romance", "action", "childrens"});
  // Setup the list of tools that a player might be required to use.
  string *tools = ({"crowbar", "hammer"});
  string hole;
  string tool;
  
  // If there are details stored in the library, then we don't need to 
  // do any more - just give the details  back.
  if (details && sizeof (details) == 2) {
    return details;
  }
  
  // Do a task-master check to see if they notice anything.
  switch(TASKER->perform_task(player, "adventuring.perception", 120, TM_FREE ) ) {
      case AWARD :
        tell_object( player, "%^YELLOW%^%^BOLD%^You feel a little more "
          "perceptive.%^RESET%^\n" );
      case SUCCEED :
        // They do!
        
        // We pick a random category over which the hole will be noticed.
        hole = element_of (cat);      
        // We pick a random tool they need to break through.
        tool = element_of (tools);
        // We put these into an array
        details = ({hole, tool});
        // And we store that array in the library.
        LIBRARY->set_player_quest_info (player->query_name(), HOLE_QUEST, 
          details);
        break;
      default :
        // No such luck!
        hole = 0;
  }
  
  // Return the details array.  This will be 0 if it hasn't been setup, or 
  // the hole,tool array if it has.
  return details;
}